@@ -55,6 +55,10 @@ Hardware Programming Interface (HPI) for AudioScience
 #define HIF_BUFFER_ALIGNMENT	(0)
 #define STREAM_BUFFER_ALIGNMENT	(0)
 
+#define LINUX_DEV_DMA_BUFFER_ALIGNMENT	(0x100000U)
+// #define LINUX_DEV_DMA_SEGMENT_MAX_SZ	(0x100000)
+// #define LINUX_DEV_DMA_SEGMENT_BOUNDARY_MASK	(DMA_SEGMENT_MAX_SZ-1)
+
 #define HPI6700_POLL_DSP_ACK
 #endif
 
@@ -409,6 +413,7 @@ static void StreamHostBufferAllocate(
 )
 {
 	uint32_t dwCommand = phm->u.d.u.buffer.dwCommand;
+	uint32_t previous_dwSizeInBytes = phb_status->dwSizeInBytes;
 
 	HPI_InitResponse(phr, phm->wObject, phm->wFunction, 0);
 
@@ -417,44 +422,34 @@ static void StreamHostBufferAllocate(
 		/* ALLOC phase, allocate a buffer with power of 2 size,
 		   get its bus address for PCI bus mastering
 		 */
-		phm->u.d.u.buffer.dwBufferSize =
-			roundup_pow_of_two(phm->u.d.u.buffer.dwBufferSize);
-
-		if (pstream->HostBufferSize <
-			phm->u.d.u.buffer.dwBufferSize) {
-			/* Invalid size */
-			phr->wError = HPI_ERROR_INVALID_DATASIZE;
-			HPI_DEBUG_LOG2(ERROR,
-				"Buffer size requested is too big %d (must be %d or less)\n",
-				phm->u.d.u.buffer.dwBufferSize,
-				pstream->HostBufferSize);
-			return;
-		}
-
-		phm->u.d.u.buffer.dwBufferSize =
-			pstream->HostBufferSize;
-		/* return old size and allocated size,
-		   so caller can detect change */
-		phr->u.d.u.stream_info.dwDataAvailable =
-			pstream->HostBufferSize;
-		phr->u.d.u.stream_info.dwBufferSize =
-			phm->u.d.u.buffer.dwBufferSize;
 
 		if (pstream->HostBufferActive) {
-			/* Already active, no action required */
-			return;
+			phr->wError = HPI_ERROR_INVALID_OPERATION;
+			goto write_phr_and_exit;
+		}
+
+		if (!phm->u.d.u.buffer.dwBufferSize) {
+			phr->wError = HPI_ERROR_INVALID_DATASIZE;
+			goto write_phr_and_exit;
+		} else {
+			phm->u.d.u.buffer.dwBufferSize =
+				roundup_pow_of_two(phm->u.d.u.buffer.dwBufferSize);
 		}
 
 		phm->u.d.u.buffer.dwPciAddress =
 			pstream->HostBufferPhys;
 		//phm->u.d.u.buffer.dwPciAddress64_hi32 = 0;
+	}
 
-		/* get the phys addr into msg for single call alloc caller
-		 * needs to do this for split alloc (or use the same message)
-		 * return the phy address for split alloc in the respose too
-		 */
-		phr->u.d.u.stream_info.dwAuxiliaryDataAvailable =
-			pstream->HostBufferPhys;
+	if (pstream->HostBufferSize <
+		phm->u.d.u.buffer.dwBufferSize) {
+		/* Invalid size */
+		phr->wError = HPI_ERROR_INVALID_DATASIZE;
+		HPI_DEBUG_LOG2(ERROR,
+			"Buffer size requested is too big %d (must be %d or less)\n",
+			phm->u.d.u.buffer.dwBufferSize,
+			pstream->HostBufferSize);
+		goto write_phr_and_exit;
 	}
 
 	if (dwCommand == HPI_BUFFER_CMD_EXTERNAL
@@ -463,6 +458,14 @@ static void StreamHostBufferAllocate(
 		   the buffer so it can start using BBM.
 		 */
 
+		if (phm->u.d.u.buffer.dwBufferSize & (phm->u.d.u.buffer.dwBufferSize - 1)) {
+			HPI_DEBUG_LOG1(ERROR,
+				"Buffer size must be 2^N not %d\n",
+				phm->u.d.u.buffer.dwBufferSize);
+			phr->wError = HPI_ERROR_INVALID_DATASIZE;
+			goto write_phr_and_exit;
+		}
+
 		phb_status->dwSamplesProcessed = 0;
 		phb_status->dwStreamState = HPI_STATE_STOPPED;
 		phb_status->dwDspIndex = 0;
@@ -481,6 +484,28 @@ static void StreamHostBufferAllocate(
 			pstream->HostBufferActive = 1;
 		}
 	}
+
+write_phr_and_exit:
+
+	// After HwMessage() returns phr has been mangled by the DSP, we want response fields to be valid so overwrite them.
+
+	if (phr->wError == HPI_ERROR_INVALID_DATASIZE) {
+		phr->u.d.u.stream_info.dwDataAvailable =
+			pstream->HostBufferSize;
+	} else {
+		/* return old size and allocated size,
+		   so caller can detect change */
+		phr->u.d.u.stream_info.dwDataAvailable =
+			previous_dwSizeInBytes;
+	}
+	phr->u.d.u.stream_info.dwBufferSize =
+		phm->u.d.u.buffer.dwBufferSize;
+	/* get the phys addr into msg for single call alloc caller
+	 * needs to do this for split alloc (or use the same message)
+	 * return the phy address for split alloc in the respose too
+	 */
+	phr->u.d.u.stream_info.dwAuxiliaryDataAvailable =
+		pstream->HostBufferPhys;
 }
 
 static void StreamHostBufferGetInfo(
@@ -1376,6 +1401,27 @@ static void FreeInterrupts(struct hpi_adapter_obj *pao)
 	/* IRQ and MSI will be disabled by pcim_release */
 }
 
+#if (HAVE_PCI_SET_CONSISTENT_DMA_MASK == 0)
+static inline int pci_set_consistent_dma_mask(struct pci_dev *dev, u64 mask)
+{
+	return dma_set_coherent_mask(&dev->dev, mask);
+}
+#endif
+
+#if (HAVE_PCI_SET_DMA_MASK == 0)
+static inline int pci_set_dma_mask(struct pci_dev *dev, u64 mask)
+{
+	return dma_set_mask(&dev->dev, mask);
+}
+#endif
+
+#if (HAVE_DMA_SET_MIN_ALIGN_MASK == 0)
+static inline dma_set_min_align_mask(struct device *dev, unsigned int min_align_mask)
+{
+	return 0;
+}
+#endif
+
 /* Map first 3 PCI memory resources
  * Also set DMA masks to 32 bit
  */
@@ -1402,6 +1448,12 @@ static hpi_err_t SetupBars(struct hpi_adapter_obj *pao)
 		phw->bar[i].size = pci_resource_len(pci_dev, i);
 	}
 
+	i = dma_set_min_align_mask(&pci_dev->dev, LINUX_DEV_DMA_BUFFER_ALIGNMENT-1);
+	if (i) {
+		dev_err(&pci_dev->dev, "Failed to set DMA buffer alignment mask to %d\n", LINUX_DEV_DMA_BUFFER_ALIGNMENT-1);
+		return HPI_ERROR_MEMORY_ALLOC;
+	}
+
 	return 0;
 }
 
@@ -1791,9 +1843,31 @@ static hpi_err_t AllocStreamMem(struct hpi_adapter_obj *pao)
 	dwOStreamBufferSize = roundup_pow_of_two(dwMaxOStreamBW / 10);
 	dwIStreamBufferSize = roundup_pow_of_two(dwMaxIStreamBW / 10);
 
+#ifdef HPI_OS_LINUX_KERNEL
+	if (pao->os.per_stream_dma_buf_max_size) {
+		// size_t dma_max_mapping_sz = dma_max_mapping_size(&pao->os.pci_dev->dev);
+		size_t dma_max_mapping_sz = (MAX_ORDER_NR_PAGES*PAGE_SIZE);
+		uint32_t dma_buf_sz = roundup_pow_of_two(pao->os.per_stream_dma_buf_max_size);
+		dwTotalSize = dma_buf_sz * pao->outstreams +
+			dma_buf_sz * pao->instreams + STREAM_BUFFER_ALIGNMENT;
+		if (dwTotalSize > dma_max_mapping_sz) {
+			HPI_DEBUG_LOG3(ERROR,
+				"Stream buffer size (%u) requested via `per_stream_dma_buf_max_size` (%u) larger than maximum DMA alloc pages (%lu), ignoring.\n",
+				dwTotalSize, dma_buf_sz, dma_max_mapping_sz);
+		} else {
+			dwOStreamBufferSize = dma_buf_sz;
+			dwIStreamBufferSize = dma_buf_sz;
+		}
+	}
+	dwTotalSize = dwOStreamBufferSize * pao->outstreams +
+		dwIStreamBufferSize * pao->instreams;
+	dwTotalSize = max(LINUX_DEV_DMA_BUFFER_ALIGNMENT, dwTotalSize);
+
+#else
 	dwTotalSize = dwOStreamBufferSize * pao->outstreams +
 		dwIStreamBufferSize * pao->instreams +
 		STREAM_BUFFER_ALIGNMENT;
+#endif
 
 	pVirtAddr = HpiOs_LockedMem_Alloc(&phw->hStreamMem,
 			dwTotalSize, &pao->os, &dwPhysAddr);
